// All database related API calls

import { useAppDispatch, useAuthorizationCache, useSessionCache } from "../store";
import { updateCurrentDatabase, updateDatabaseList } from "../store/sessionSlice";

export enum DatabaseType {
  ArangoDB = 0,
  Neo4j = 1,
}

export const databaseNameMapping: string[] = ['arangodb', 'neo4j'];

export type AddDatabaseRequest = {
  name: string;
  internal_database_name: string;
  url: string;
  port: number;
  username: string;
  password: string;
  type: DatabaseType; // Database type. 0 = ArangoDB, 1 = Neo4j
};

export type AddDatabaseOptions = {
  setAsCurrent?: boolean, updateDatabaseCache?: boolean
}

export type GetDatabasesOptions = {
  updateSessionCache?: boolean
}

export type DeleteDatabasesOptions = {
  updateSessionCache?: boolean
}

export const useDatabaseAPI = (domain: string) => {
  const { accessToken } = useAuthorizationCache();
  const cache = useSessionCache();
  const dispatch = useAppDispatch();

  function AddDatabase(
    request: AddDatabaseRequest,
    options: AddDatabaseOptions = {}
  ): Promise<void> {
    const { setAsCurrent = true, updateDatabaseCache = false } = options;
    return new Promise((resolve, reject) => {
      fetch(`https://${domain}/user/database`, {
        method: 'POST',
        credentials: 'same-origin',
        headers: new Headers({
          Authorization:
            'Bearer ' + accessToken,
        }),
        body: JSON.stringify(request),
      }).then((response: Response) => {
        console.log(response);

        if (!response.ok) {
          reject(response.statusText);
        }
        if (setAsCurrent)
          dispatch(updateCurrentDatabase(request.name));
        if (updateDatabaseCache)
          GetAllDatabases({ updateSessionCache: true });

        resolve();
      });
    });
  }

  async function GetAllDatabases(options: GetDatabasesOptions = {}): Promise<Array<string>> {
    const { updateSessionCache: updateDatabaseCache = true } = options;
    // console.log(accessToken);
    const response = await fetch(`https://${domain}/user/database`, {
      method: 'GET',
      credentials: 'same-origin',
      headers: new Headers({
        Authorization:
          'Bearer ' + accessToken,
      }),
    });
    console.log(response);

    if (!response.ok) {
      const text = await response.text();
      console.error(text);
      return [];
      // throw Error(response.statusText)
    }
    const json = await response.json();
    if (updateDatabaseCache)
      dispatch(updateDatabaseList(json.databases));
    return json.databases;
  }

  function DeleteDatabase(name: string, options: DeleteDatabasesOptions = {}): Promise<void> {
    const { updateSessionCache: updateDatabaseCache = true } = options;
    return new Promise((resolve, reject) => {
      fetch(`https://${domain}/user/database/` + name, {
        method: 'DELETE',
        credentials: 'same-origin',
        headers: new Headers({
          Authorization:
            'Bearer ' + accessToken,
        }),
      }).then((response: Response) => {
        if (!response.ok) {
          reject(response.statusText);
        }

        if (updateDatabaseCache)
          GetAllDatabases({ updateSessionCache: true });

        resolve();
      });
    });
  }

  return { DatabaseType, AddDatabase, GetAllDatabases, DeleteDatabase }
};