import {
  setQuerybuilderNodes,
  useAppDispatch,
  useConfig,
  useQuerybuilderGraph,
  useQuerybuilderGraphology,
} from '@graphpolaris/shared/lib/data-access/store';
import ReactFlow, {
  ReactFlowProvider,
  Background,
  Node,
  isNode,
  ReactFlowInstance,
  Controls,
  ControlButton,
  NodeChange,
  ConnectionMode,
} from 'reactflow';
import styles from './querybuilder.module.scss';

import React, { ReactComponentElement, useMemo, useRef } from 'react';
import {
  AttributePill,
  ConnectionDragLine,
  ConnectionLine,
  EntityFlowElement,
  RelationPill,
} from '../pills';
import { createReactFlowElements } from '../graph/reactflow/utils';
import {
  dragPillStarted,
  dragPillStopped,
  movePillTo,
} from '../pills/dragging/dragPill';
import {
  Settings as SettingsIcon,
  Delete as DeleteIcon,
  ImportExport as ExportIcon,
} from '@mui/icons-material';
import { clearQB } from '@graphpolaris/shared/lib/data-access/store/querybuilderSlice';
import {
  EntityNode,
  QueryElementTypes,
  RelationNode,
} from '@graphpolaris/shared/lib/querybuilder/graph/reactflow/model';
import {
  RelationPosToFromEntityPos,
  RelationPosToToEntityPos,
} from '@graphpolaris/shared/lib/querybuilder/graph/graphology/utils';
import { Handles } from '@graphpolaris/shared/lib/querybuilder/graph/reactflow/handles';
import { useDispatch } from 'react-redux';

const nodeTypes = {
  entity: EntityFlowElement,
  relation: RelationPill,
  attribute: AttributePill,
};
const edgeTypes = {
  connection: ConnectionLine,
  attribute_connection: ConnectionLine,
};

/**
 * This is the main querybuilder component. It is responsible for holding all pills and fire off the visual part of the querybuilder panel logic
 */
export const QueryBuilder: React.FC = () => {
  const graphologyGraph = useQuerybuilderGraphology();
  const graph = useQuerybuilderGraph();
  const config = useConfig();
  const dispatch = useDispatch();
  const isDraggingPill = useRef(false);
  // console.log('inputnodes', nodes);
  const reactFlowInstanceRef = useRef<ReactFlowInstance>();
  const divRef = useRef<HTMLDivElement>(null);

  const onInit = (reactFlowInstance: ReactFlowInstance) => {
    reactFlowInstanceRef.current = reactFlowInstance;
    setTimeout(() => reactFlowInstance.fitView(), 0);
  };

  const elements = useMemo(
    () => createReactFlowElements(graphologyGraph),
    [graphologyGraph]
  );

  /**
   * Called when a node is dragged in querybuilder to allow for movement
   * @param event
   * @param node
   */
  const onNodeDrag = (
    event: React.MouseEvent<Element, MouseEvent>,
    node: Node<any>
  ) => {
    // Get the node in the elements list to get the previous location
    const pNode = elements.nodes.find((e) => e?.id === node?.id);
    if (!(pNode && isNode(pNode))) return;
    // This is then used to calculate the delta position
    const dx = node.position.x - pNode.position.x;
    const dy = node.position.y - pNode.position.y;

    // console.log(node, pNode);
    // Check if we started dragging, if so, call the drag started usecase
    if (!isDraggingPill.current) {
      dragPillStarted(node.id, graphologyGraph);
      isDraggingPill.current = true;
    }

    // Call the drag usecase
    movePillTo(node.id, graphologyGraph, dx, dy, node.position);

    // Dispatch the new graphology object, so reactflow will get rerendered
    dispatch(setQuerybuilderNodes(graphologyGraph.export()));
  };

  /**
   * Called when a node is released from dragging in querybuilder
   * @param event
   * @param node
   **/
  const onNodeDragStop = (
    event: React.MouseEvent<Element, MouseEvent>,
    node: Node<any>
  ) => {
    // isDraggingPill.current = false;
    //
    // // Call the drag pill stopped usecase
    // dragPillStopped(node.id, graphologyGraph);
    //
    // // Dispatch the new graphology object, so reactflow will get rerendered
    // dispatch(setQuerybuilderNodes(graphologyGraph.export()));
  };

  /**
   * Clears all nodes in the graph.
   */
  function clearAllNodes() {
    dispatch(clearQB());
  }

  /**
   * Clears all nodes in the graph.
   *  TODO: only works if the node is clicked and not moved (maybe use onSelectionChange)
   */
  function onNodesDelete(nodes: Node[]) {
    nodes.forEach((n) => {
      graphologyGraph.dropNode(n.id);
    });

    // Dispatch the new graphology object, so reactflow will get re-rendered
    dispatch(setQuerybuilderNodes(graphologyGraph.export()));
  }

  /**
   * TODO?
   */
  function onNodesChange(nodes: NodeChange[]) {
    // console.log(nodes);
    nodes.forEach((n) => {
      // console.log(graphologyGraph.findNode((gn) => gn === n?.id));
      if (n.type !== 'position') {
        // updated = true;
        // graphologyGraph.updateAttributes(n.id, n.data);
      }
    });
  }

  const onDragOver = (event: React.DragEvent<HTMLDivElement>): void => {
    event.preventDefault();
    event.dataTransfer.dropEffect = 'move';
  };

  /**
   * The onDrop is called when the user drops an element from the schema onto the QueryBuilder.
   * In the onDrop query elements will be created based on the data stored in the drag event (datastrasfer).
   * @param event Drag event.
   */
  const onDrop = (event: React.DragEvent<HTMLDivElement>): void => {
    event.preventDefault();

    // The dropped element should be a valid reactflow element
    const data: string = event.dataTransfer.getData('application/reactflow');
    if (data.length == 0 || !reactFlowInstanceRef?.current) return;

    const dragData = JSON.parse(data);
    const reactFlow = divRef.current as HTMLDivElement;
    const reactFlowBounds = reactFlow.getBoundingClientRect();
    const position = reactFlowInstanceRef.current.project({
      //TODO: this position should be centre of entity, rather than topleft
      x: event.clientX - reactFlowBounds.left,
      y: event.clientY - reactFlowBounds.top,
    });

    switch (dragData.type) {
      case QueryElementTypes.Entity:
        graphologyGraph.addPill2Graphology({
          type: QueryElementTypes.Entity,
          x: position.x,
          y: position.y,
          name: dragData.name,
          fadeIn: dragData.fadeIn,
        });
        break;
      // Creates a relation element and will also create the 2 related entities together with the connections
      case QueryElementTypes.Relation:
        const relationId = graphologyGraph.addPill2Graphology({
          type: QueryElementTypes.Relation,
          x: position.x,
          y: position.y,
          depth: { min: 0, max: 1 },
          // name: dragData.name,
          name: dragData.collection, // TODO leave collection or use name?
          fadeIn: dragData.fadeIn,
          collection: dragData.collection,
        });
        const leftEntityId = graphologyGraph.addPill2Graphology({
          type: QueryElementTypes.Entity,
          ...RelationPosToFromEntityPos(position),
          name: dragData.from,
          fadeIn: true,
        });
        const rightEntityId = graphologyGraph.addPill2Graphology({
          type: QueryElementTypes.Entity,
          ...RelationPosToToEntityPos(position),
          name: dragData.to,
          fadeIn: true,
        });

        graphologyGraph.addEdge(leftEntityId, relationId, {
          type: 'connection',
          sourceHandle: Handles.ToRelation,
          targetHandle: Handles.RelationLeft,
        });
        graphologyGraph.addEdge(rightEntityId, relationId, {
          type: 'connection',
          sourceHandle: Handles.ToRelation,
          targetHandle: Handles.RelationRight,
        });

        if (config.autoSendQueries) {
          // sendQuery();
        }

        dispatch(setQuerybuilderNodes(graphologyGraph.export()));
        break;
      // Creates an attribute element with the correct dataType
      // case QueryElementTypes.Attribute:
      //   createNodeFromSchema(
      //     QueryElementTypes.Attribute,
      //     position,
      //     dragData.name,
      //     dragData.fadeIn,
      //     dragData.datatype
      //   );
      //   break;
    }
  };

  return (
    <div
      style={{
        width: '100%',
        height: '100%',
      }}
      ref={divRef}
    >
      <ReactFlowProvider>
        <ReactFlow
          edges={elements.edges}
          nodes={elements.nodes}
          snapGrid={[10, 10]}
          snapToGrid
          nodeTypes={nodeTypes}
          edgeTypes={edgeTypes}
          connectionLineComponent={ConnectionDragLine}
          // connectionMode={ConnectionMode.Loose}
          onInit={onInit}
          onNodeDrag={onNodeDrag}
          onNodeDragStop={onNodeDragStop}
          onDragOver={onDragOver}
          // onConnect={this.queryBuilderViewModel.onConnect}
          onDrop={onDrop}
          onNodesDelete={onNodesDelete}
          onNodesChange={onNodesChange}
          deleteKeyCode="Backspace"
          className={styles.reactflow}
          attributionPosition="top-right"
        >
          <Background gap={10} size={0.7} />
          <Controls
            showZoom={false}
            showInteractive={false}
            className={styles.controls}
          >
            <ControlButton
              className={styles.buttons}
              title={'Remove all elements'}
              onClick={() => clearAllNodes()}
            >
              <DeleteIcon />
            </ControlButton>
            <ControlButton
              className={styles.buttons}
              title={'Export querybuilder'}
              onClick={(event) => {
                event.stopPropagation();
                // this.setState({
                //   ...this.state,
                //   exportMenuAnchor: event.currentTarget,
                // });
              }}
            >
              <ExportIcon />
            </ControlButton>
            <ControlButton
              className={styles.buttons}
              title={'Other settings'}
              onClick={(event) => {
                event.stopPropagation();
                // this.setState({
                //   ...this.state,
                //   settingsMenuAnchor: event.currentTarget,
                // });
              }}
            >
              <SettingsIcon />
            </ControlButton>
          </Controls>
        </ReactFlow>
      </ReactFlowProvider>
    </div>
  );
};

export default QueryBuilder;
